from __future__ import annotations

import logging
import os
import time as ttime

import arrow
import dask.array as da
import numpy as np
import scipy as sp
from jax import jit
from jax import scipy as jsp
from tqdm import tqdm

from ..beam import compute_angular_fwhm
from ..constants import k_B
from ..io import DEFAULT_BAR_FORMAT, DEFAULT_TIME_FORMAT, fetch, humanize_time
from ..map import HEALPixMap, Map, ProjectionMap, load
from ..units import Quantity

here, this_filename = os.path.split(__file__)
logger = logging.getLogger("maria")


class MapMixin:
    """
    This simulates scanning over celestial sources.

    TODO: add errors
    """

    def _initialize_map(self, map: str | Map, **map_kwargs):
        if isinstance(map, str):
            self.map = load(fetch(map), **map_kwargs)
        elif isinstance(map, Map):
            self.map = map
        else:
            raise ValueError("'map' must be either a Map or a string")

        # the map can be frequency-naive if it is already in K_RJ
        # self.map = map.to(units="K_RJ")

        if "stokes" not in self.map.dims:
            self.map = self.map.unsqueeze("stokes")

        if "nu" not in self.map.dims:
            self.map = self.map.unsqueeze("nu")

        if "t" in self.map.dims:
            if self.map.dims["t"] > 1:
                map_start = arrow.get(self.map.t.min()).to("utc")
                map_end = arrow.get(self.map.t.max()).to("utc")
                if map_start > self.min_time:
                    logger.warning(
                        f"Beginning of map ({map_start.format(DEFAULT_TIME_FORMAT)}) is after the "
                        f"beginning of the simulation ({self.min_time.format(DEFAULT_TIME_FORMAT)}).",
                    )
                if map_end < self.max_time:
                    logger.warning(
                        f"End of map ({map_end.format(DEFAULT_TIME_FORMAT)}) is before the "
                        f"end of the simulation ({self.max_time.format(DEFAULT_TIME_FORMAT)}).",
                    )
        else:
            self.map = self.map.unsqueeze("t")

    def _run(self, **kwargs):
        self._sample_maps(**kwargs)

    def _sample_maps(self, obs):
        map_loading = np.zeros(obs.coords.shape, dtype=self.dtype)

        bands_pbar = tqdm(
            obs.instrument.dets.bands,
            desc="Sampling map",
            disable=self.disable_progress_bars,
            bar_format=DEFAULT_BAR_FORMAT,
            ncols=250,
            postfix={"band": "", "channel": "", "stokes": ""},
        )
        for band in bands_pbar:
            bands_pbar.set_postfix(band=band.name)

            band_mask = obs.instrument.dets.band_name == band.name
            band_coords = obs.coords[band_mask]
            band_dets = obs.instrument.dets[band_mask]

            pointing_s = ttime.monotonic()
            P = self.map.pointing_matrix(coords=band_coords, dets=band_dets)
            logger.debug(f"Computed pointing matrix for band {band.name} in {humanize_time(ttime.monotonic() - pointing_s)}")

            band_fwhm = Quantity(
                compute_angular_fwhm(
                    fwhm_0=obs.instrument.dets.primary_size.mean(),
                    z=np.inf,
                    nu=band.center.Hz,
                ),
                "rad",
            )

            # ideally we would do this for each nu bin, but that's slow
            smoothed_map = self.map.smooth(fwhm=band_fwhm)

            for channel_index, (nu_min, nu_max) in enumerate(self.map.nu_bin_bounds):
                channel_s = ttime.monotonic()
                channel_map = smoothed_map.to("K_RJ", band=band).data[:, channel_index]
                qchannel = (nu_min, nu_max)
                channel_string = f"{qchannel}"

                spectrum_kwargs = (
                    {
                        "spectrum": obs.atmosphere.spectrum,
                        "zenith_pwv": obs.zenith_scaled_pwv[band_mask].compute(),
                        "base_temperature": obs.atmosphere.weather.temperature[0],
                        "elevation": obs.coords.el[band_mask],
                    }
                    if getattr(obs, "atmosphere", None)
                    else {}
                )

                calibration_s = ttime.monotonic()
                pW_per_K_RJ = (
                    1e12
                    * k_B
                    * band.compute_transmission_integral(nu_min_Hz=nu_min.Hz, nu_max_Hz=nu_max.Hz, **spectrum_kwargs)
                )
                logger.debug(
                    f"Computed K_RJ -> pW calibration for band {band.name}, channel {channel_string} in "
                    f"{humanize_time(ttime.monotonic() - calibration_s)}"
                )

                # for stokes_index, stokes in enumerate(getattr(self.map, "stokes", "I")):
                #    bands_pbar.set_postfix(band=band.name, channel=channel_string, stokes=stokes)

                # stokes_weight = stokes_weights[band_mask, "IQUV".index(stokes), None]
                # if np.isclose(stokes_weight, 0).all():
                #     logger.debug(f"Skipping stokes {stokes} (no weight)")
                #     continue

                # stokes_channel_s = ttime.monotonic()
                # channel_stokes_map = channel_map[stokes_index].compute()

                # sample_T_RJ = jit(jsp.interpolate.RegularGridInterpolator(
                #             (self.map.y_side[::-1], self.map.x_side),
                #             channel_stokes_map[::-1],
                #             bounds_error=False,
                #             fill_value=0,
                #             method="linear",
                #         ))((dy[band_mask], dx[band_mask]))

                # flat_padded_map = np.pad(channel_stokes_map, pad_width=((1, 1)), mode="edge").ravel()
                # flat_padded_map = np.pad(channel_stokes_map, pad_width=((1, 1)), mode="edge").ravel()

                pW = pW_per_K_RJ * (P @ channel_map.ravel()).reshape(band_coords.shape)

                map_loading[band_mask] += pW

                if logger.level == 10:
                    logger.debug(
                        f"Computed map load {pW.shape} ~{Quantity(pW.std(), 'pW'):<8} for band {band.name}, "
                        f"channel {channel_string} in {humanize_time(ttime.monotonic() - channel_s)}"
                    )

                del pW

            if map_loading[band_mask].sum() == 0:
                logger.warning(f"No load from map for band {band.name}")

        obs.loading["map"] = da.asarray(map_loading, dtype=self.dtype)
        del map_loading



from __future__ import annotations

import logging
import warnings
from itertools import cycle

import arrow
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import scipy as sp
from matplotlib.animation import FuncAnimation
from matplotlib.collections import EllipseCollection
from matplotlib.patches import Patch

from maria.beam import compute_angular_fwhm
from maria.units import Quantity, parse_units
from maria.utils.signal import detrend as detrend_signal

logger = logging.getLogger("maria")

FIELD_LABELS = {"atmosphere": "atm."}


def plot_tod(
    tod,
    detrend: str = "mean",
    n_freq_bins: int = 1024,
    max_dets: int = 1,
    lw: float = 1e0,
    fontsize: float = 10,
    figsize: tuple[float, float] = (10, 5),
):
    fig, axes = plt.subplots(
        ncols=2,
        nrows=len(tod.fields),
        sharex="col",
        figsize=figsize,
        dpi=256,
        constrained_layout=True,
    )
    axes = np.atleast_2d(axes)
    gs = axes[0, 0].get_gridspec()

    # plt.subplots_adjust(top=0.99, bottom=0.01, hspace=0, wspace=0.025)

    for ax in axes[:, -1]:
        ax.remove()
    ps_ax = fig.add_subplot(gs[:, -1])

    handles = []
    colors = cycle(plt.rcParams["axes.prop_cycle"].by_key()["color"])

    tod_ubands = np.unique(tod.dets.band_name)
    power_spectra = np.zeros((len(tod_ubands), len(tod.fields), n_freq_bins - 1))

    tod_u = parse_units(tod.units)

    for field_index, field in enumerate(tod.fields):
        data = tod.data[field]

        tod_ax = axes[field_index, 0]
        field_label = FIELD_LABELS.get(field, field)

        field_plot_data = {}

        for band_index, band_name in enumerate(tod_ubands):
            band_mask = tod.dets.band_name == band_name

            field_plot_data[band_name] = data[band_mask]

            if band_mask.sum() > max_dets:
                field_plot_data[band_name] = field_plot_data[band_name][
                    np.linspace(0, band_mask.sum() - 1, max_dets).astype(int)
                ]

            if detrend == "mean":
                field_plot_data[band_name] = field_plot_data[band_name] - field_plot_data[band_name].mean(axis=-1)[..., None]

            if detrend == "linear":
                field_plot_data[band_name] = detrend_signal(field_plot_data[band_name], order=1)

        max_abs_signal = Quantity(
            [np.abs(field_plot_data[band_name]).max().compute() for band_name in tod_ubands], units=tod.units
        ).max()
        field_u = max_abs_signal.u
        logger.debug(
            f"inferring units '{max_abs_signal.u['units']}' for field '{field}' from max abs signal '{max_abs_signal}'"
        )

        for band_index, band_name in enumerate(tod_ubands):
            color = next(colors)

            band_mask = tod.dets.band_name == band_name
            f, ps = sp.signal.periodogram(
                detrend_signal(tod.data[field][band_mask].compute(), order=1), fs=tod.fs, window="hann"
            )

            f_bins = np.geomspace(0.999 * f[1], 1.001 * f[-1], n_freq_bins)
            f_mids = np.sqrt(f_bins[1:] * f_bins[:-1])

            binned_ps = sp.stats.binned_statistic(
                f,
                ps.mean(axis=0),
                bins=f_bins,
                statistic="mean",
            )[0]

            power_spectra[band_index, field_index] = binned_ps

            use = binned_ps > 0

            ps_ax.plot(
                f_mids[use],
                binned_ps[use],
                lw=5e-1,
                color=color,
            )
            tod_ax.plot(
                tod.time,
                Quantity(field_plot_data[band_name].T, units=tod.units).to(field_u["units"]),
                lw=5e-1,
                color=color,
                zorder=-band_index,
            )

            handles.append(Patch(color=color, label=f"{field_label} ({band_name}, $n_\\text{{det}} = {band_mask.sum()}$)"))

        tod_ax.set_xlim(tod.time.min(), tod.time.max() + 1e-1)
        tod_ax.set_ylabel(f"{field_label} [${field_u['math_name']}$]", fontsize=fontsize)

        if field_index == 0:
            time_label_ax = tod_ax.twiny()

    tod_ax.set_xlabel("Timestamp [s]", fontsize=fontsize)

    time_label_ticks = np.linspace(tod.time.min(), tod.time.max(), 3)
    time_label_ticks -= time_label_ticks % 60
    time_label_ax.set_xticks(np.unique(time_label_ticks))
    time_label_ax.set_xticklabels([arrow.get(t).strftime("%Y-%m-%d\n%H:%M:%S") for t in time_label_ax.get_xticks()])
    time_label_ax.set_xlim(tod.time.min(), tod.time.max() + 1e-1)

    ps_ax.yaxis.tick_right()
    ps_ax.yaxis.set_label_position("right")
    ps_ax.set_xlabel("Frequency [Hz]", fontsize=fontsize)
    ps_ax.set_ylabel(f"Power spectral density [${tod_u['math_name']}^2$/Hz]", fontsize=fontsize)
    ps_ax.legend(handles=handles, loc="upper right", fontsize=0.8 * fontsize)
    ps_ax.loglog()
    ps_ax.set_xlim(f_mids.min(), f_mids.max())

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        # I want to see at least the top order of magnitude of all fields
        minimum_dominant_ps = np.nanmin(np.nanmax(power_spectra, axis=1))
        minimum_maximum_ps = np.nanmin(np.nanmax(power_spectra, axis=2))

    ps_ax.set_ylim(1e-1 * np.minimum(minimum_dominant_ps, minimum_maximum_ps))


def twinkle_plot(tod, rate=2, fps=30, start_index=0, max_frames=100, filename=None):
    fps = np.minimum(fps, tod.fs)
    time_offset = tod.time.min()

    frame_time = np.arange(tod.time.min(), tod.time.max(), rate / fps)[:max_frames]
    frame_index = np.interp(frame_time, tod.time, np.arange(len(tod.time))).astype(int)

    offsets = Quantity(np.c_[tod.dets.sky_x, tod.dets.sky_y], "rad")
    fwhms = Quantity(compute_angular_fwhm(fwhm_0=tod.dets.primary_size, nu=tod.dets.band_center), "rad")

    bands = sorted(np.unique(tod.dets.band_name))

    nrows = int(np.sqrt(len(bands)))
    ncols = int(np.ceil(len(bands) / nrows))
    subplot_width = np.minimum(6, 10 / nrows)
    subplot_height = np.minimum(6, 10 / ncols)
    subplot_size = np.minimum(subplot_width, subplot_height)
    figure_size = (ncols * subplot_size, nrows * (subplot_size + 1))

    fig, axes = plt.subplots(
        nrows=nrows,
        ncols=ncols,
        sharex=True,
        sharey=True,
        figsize=figure_size,
        dpi=256,
        constrained_layout=True,
    )

    axes_iterator = iter(np.atleast_1d(axes).ravel())

    subplots = {}

    for band in bands:
        ax = next(axes_iterator)

        band_mask = tod.dets.band_name == band

        ec = EllipseCollection(
            widths=getattr(fwhms, offsets.units)[band_mask],
            heights=getattr(fwhms, offsets.units)[band_mask],
            angles=0,
            units="xy",
            offsets=getattr(offsets, offsets.units)[band_mask],
            transOffset=ax.transData,
            cmap="cmb",
        )

        ax.add_collection(ec)
        ax.scatter(
            *getattr(offsets, offsets.units)[band_mask].T,
            label=band,
            s=0,
            color="k",
        )

        ax.set_xlabel(r"$\Delta \theta_x$ [deg.]")
        ax.set_ylabel(r"$\Delta \theta_y$ [deg.]")

        boresight_info = ax.text(
            0.01,
            0.99,
            "",
            ha="left",
            va="top",
            transform=ax.transAxes,
        )

        band_qdata = Quantity(tod.signal[band_mask][..., frame_index].compute(), units=tod.units)

        cbar = fig.colorbar(ec, ax=ax, shrink=0.8, location="bottom")

        cbar.set_label(f"{band_qdata.q['long_name']} [${band_qdata.u['math_name']}$]")

        subplots[band] = {
            "ax": ax,
            "ec": ec,
            "cbar": cbar,
            "info": boresight_info,
            "time": tod.time[frame_index] - time_offset,
            "az": tod.boresight.az[frame_index],
            "el": tod.boresight.el[frame_index],
            "data": band_qdata.value,
        }

    for ax in axes_iterator:
        ax.set_axis_off()

    def update(frame, subplots):
        # i = frame_index[frame]

        for band, subplot in subplots.items():
            info = f"time = {time_offset:.00f} + {subplot['time'][frame]:.02f} s\naz = {np.degrees(subplot['az'][frame]):.02f} deg\nel = {np.degrees(subplot['el'][frame]):.02f} deg"  # noqa

            buffer = 4

            norm_start = np.maximum(0, frame - buffer)
            norm_end = np.minimum(len(frame_index) - 1, frame + buffer)

            # if frame < buffer:
            #     norm_start, norm_end = 0, 2 * buffer
            # if frame > len(t) - buffer:
            #     norm_start, norm_end = len(t) - 2 * buffer - 1, len(t) - 1

            norm = mpl.colors.Normalize(
                *np.quantile(subplot["data"][:, norm_start:norm_end], q=[0.01, 0.99]),
            )

            subplot["ec"].set_array(subplot["data"][:, frame])
            subplot["ec"].set_norm(norm)
            subplot["cbar"].update_normal(subplot["ec"])
            subplot["info"].set_text(info)  # noqa

        return subplots

    update(0, subplots)

    ani = FuncAnimation(
        fig=fig,
        func=update,
        fargs=(subplots,),
        frames=len(frame_index),
        interval=1e3 / fps,
    )

    if filename:
        ani.save(filename)



from __future__ import annotations

import inspect
import logging

import numpy as np
import pandas as pd
import scipy as sp

logger = logging.getLogger("maria")

VALID_SCAN_KWARGS = [
    "time",
    "radius",
    "width",
    "height",
    "x_throw",
    "y_throw",
    "speed",
    "n",
    "petals",
    "ratio",
    "freq_ratio",
    "miss_factor",
    "smoothness",
]


def parse_scan_kwargs(scan_kwargs, default_radius: float = 1.0):
    for kwarg in scan_kwargs:
        if kwarg not in VALID_SCAN_KWARGS:
            raise ValueError(f"Invalid scan kwarg '{kwarg}'")

    size_kwargs = ["radius", "width", "x_throw", "height", "y_throw"]
    if not any([kwarg in scan_kwargs for kwarg in size_kwargs]):
        if default_radius is None:
            default_radius = 1.0
            logger.warning(
                f"No scan size kwargs (one of {size_kwargs}) were passed. "
                f"Assuming a scan radius of {default_radius:.03e} degrees."
            )
        scan_kwargs["radius"] = default_radius

    if "x_throw" not in scan_kwargs:
        if "radius" in scan_kwargs:
            scan_kwargs["x_throw"] = scan_kwargs.pop("radius")
        elif "width" in scan_kwargs:
            scan_kwargs["x_throw"] = 0.5 * scan_kwargs.pop("width")
        elif "y_throw" in scan_kwargs:
            scan_kwargs["x_throw"] = scan_kwargs["y_throw"]
        else:
            scan_kwargs["x_throw"] = 0.5 * scan_kwargs.pop("height")

    if "y_throw" not in scan_kwargs:
        if "height" in scan_kwargs:
            scan_kwargs["y_throw"] = 0.5 * scan_kwargs.pop("height")
        else:
            scan_kwargs["y_throw"] = scan_kwargs["x_throw"]

    if "speed" not in scan_kwargs:
        scan_kwargs["speed"] = max(scan_kwargs["x_throw"], scan_kwargs["y_throw"]) / 4

    logger.debug(f"Parsed scan pattern kwargs {scan_kwargs}")

    return scan_kwargs


def lissajous(
    time,
    x_throw,
    y_throw,
    speed,
    freq_ratio=1.193,
    **extra_kwargs,
):  # noqa
    if extra_kwargs:
        logger.warning(f"Ignoring parameters {extra_kwargs} for scan pattern 'lissajous'.")

    freq = speed / np.sqrt((x_throw * freq_ratio) ** 2 + y_throw**2)

    x = x_throw * np.cos(freq_ratio * freq * time)
    y = y_throw * np.sin(freq * time)

    return np.stack([x, y])


def double_circle(time, x_throw, y_throw, speed, ratio=0.5, freq_ratio=1.7, **extra_kwargs):
    if extra_kwargs:
        logger.warning(f"Ignoring parameters {extra_kwargs} for scan pattern 'double_circle'.")

    radius = x_throw

    a = radius / (1 + 1 / ratio)
    b = a / ratio

    phase = time * speed / np.maximum(a + b * freq_ratio, 1e-16)  # do not divide by zero!

    x = a * np.sin(phase) + b * np.sin(phase * freq_ratio)
    y = a * np.cos(phase) + b * np.cos(phase * freq_ratio)

    return np.stack([x, (y_throw / x_throw) * y])


def daisy_from_phase(phase, a, b, petals, miss_freq):
    x_p = a * np.cos(petals * phase) * np.sin(phase) + b * np.sin(petals * phase) * np.cos(miss_freq * phase)
    y_p = a * np.cos(petals * phase) * np.cos(phase) + b * np.sin(petals * phase) * np.sin(miss_freq * phase)
    X = np.stack([x_p, y_p])
    return (a + b) * X / np.sqrt(np.square(X).sum(axis=0).max())


def daisy(
    time,
    x_throw,
    y_throw,
    speed,
    petals=7 / np.pi,
    miss_factor=0.15,
    miss_freq=np.sqrt(2),
    **extra_kwargs,
):  # noqa
    if extra_kwargs:
        logger.warning(f"Ignoring parameters {extra_kwargs} for scan pattern 'daisy'.")

    radius = x_throw

    if radius > 0:
        a = radius / (1 + miss_factor)
        b = a * miss_factor
        max_speed = 0.0
        dp_dt = (speed / radius) if radius > 0.0 else 0.0

        for _ in range(4):
            phase = dp_dt * time
            test_x, test_y = daisy_from_phase(phase, a=a, b=b, petals=petals, miss_freq=miss_freq)
            vx2 = (np.gradient(test_x) / np.gradient(time)) ** 2
            vy2 = (np.gradient(test_y) / np.gradient(time)) ** 2
            max_speed = np.sqrt(vx2 + vy2).max()

            if np.abs(np.log(max_speed / speed)) > 0.01:
                dp_dt *= speed / max_speed
            else:
                break

        x, y = daisy_from_phase(phase, a=a, b=b, petals=petals, miss_freq=miss_freq)

    else:
        x, y = np.zeros((2, len(time)))

    return np.stack([x, (y_throw / x_throw) * y])


def smooth_sawtooth(p, delta=0.01):
    return 1 - 2 * np.arccos((delta - 1) * np.cos(p)) / np.pi


def back_and_forth(t, radius=1, x_throw=None, y_throw=0, speed=1.0, max_accel=np.inf, d=0.01):
    x_throw = x_throw or radius

    factor = 1 / (1 - 2 * np.arccos(1 - d) / np.pi)

    throw = factor * np.sqrt(x_throw**2 + y_throw**2)

    a = np.pi * speed / (2 * throw * (1 - d))
    b = np.sqrt(np.pi * max_accel * np.sqrt(2 * d - d**2) / (2 * throw * (1 - d)))

    dp_dt = np.minimum(a, b)

    x = factor * x_throw * smooth_sawtooth(dp_dt * t, delta=d)
    y = factor * y_throw * smooth_sawtooth(dp_dt * t, delta=d)

    return np.stack([x, y])


def raster(time: float, x_throw: float, y_throw: float, speed: float, n: int = 21, smoothness: float = 2, **extra_kwargs):
    if extra_kwargs:
        logger.warning(f"Ignoring parameters {extra_kwargs} for scan pattern 'raster'.")

    sample_rate = 1 / np.median(np.gradient(time))

    raster_distance = np.sqrt((2 * n * x_throw) ** 2 + (2 * y_throw) ** 2)
    raster_duration = raster_distance / speed

    raster_phase = np.linspace(0, 2 * np.pi, 10000)
    raster_x = x_throw * sp.signal.sawtooth(0.5 * n * raster_phase, width=0.5)
    raster_y = y_throw * np.linspace(1, -1, 10000)

    return_distance = np.sqrt((raster_x[-1] + x_throw) ** 2 + (2 * y_throw) ** 2)
    return_duration = return_distance / speed
    period = raster_duration + return_duration

    raster_phase *= raster_duration / period

    phase_samples = [*raster_phase, 2 * np.pi]
    offset_samples = np.stack([[*raster_x, -x_throw], [*raster_y, y_throw]], axis=0)

    phase = 2 * np.pi * (time % period) / period

    offsets = sp.interpolate.interp1d(phase_samples, offset_samples)(phase)

    return sp.ndimage.gaussian_filter1d(offsets, sigma=1, axis=-1)


def stare(time, **extra_kwargs):
    if extra_kwargs:
        logger.warning(f"Ignoring parameters {extra_kwargs} for scan pattern 'stare'.")
    return np.zeros((2, *time.shape))


# def get_constant_speed_offsets(
#     pattern,
#     duration,
#     sample_rate,
#     speed,
#     eps=1e-6,
#     **scan_options,
# ):
#     """This should be cythonized maybe."""

#     speed_per_sample = speed / sample_rate

#     def phase_coroutine():
#         p = 0.0

#         for _ in range(int(duration * sample_rate)):
#             x0, y0 = pattern(p - 0.5 * eps, **scan_options)
#             x1, y1 = pattern(p + 0.5 * eps, **scan_options)

#             ds_dp = np.sqrt((x1 - x0) ** 2 + (y1 - y0) ** 2) / eps  # speed per phase

#             dp = speed_per_sample / ds_dp
#             p += dp

#             yield p

#     return pattern(np.array([p for p in phase_coroutine()]), **scan_options)


scan_patterns = {
    "stare": {"aliases": [], "generator": stare},
    "daisy": {"aliases": ["daisy_scan"], "generator": daisy},
    "lissajous": {"aliases": ["lissajous_box"], "generator": lissajous},
    "raster": {"aliases": [], "generator": raster},
    "back_and_forth": {"aliases": ["back-and-forth"], "generator": back_and_forth},
    # "grid": {"aliases": [], "generator": grid},
    "double_circle": {"aliases": [], "generator": double_circle},
}

for key in scan_patterns:
    scan_patterns[key]["signature"] = str(inspect.signature(scan_patterns[key]["generator"]))

scan_patterns = pd.DataFrame(scan_patterns).T.sort_index()


def get_scan_pattern_generator(pattern):
    for index, entry in scan_patterns.iterrows():
        if (pattern == index) or (pattern in entry.aliases):
            return entry.generator

    raise ValueError(f"Invalid scan pattern '{pattern}'. Valid scan patterns are {list(scan_patterns.index)}.")


def generate_scan_offsets(time: float, pattern: str, **scan_kwargs):
    f = get_scan_pattern_generator(pattern=pattern)
    return f(time, parse_scan_kwargs(scan_kwargs))



from __future__ import annotations

import logging
import os
import warnings
from collections.abc import Sequence
from typing import Mapping

import arrow
import numpy as np
import scipy as sp
from tqdm import tqdm

from ..coords import FRAMES, Frame, infer_center_width_height
from ..instrument import BandList
from ..io import DEFAULT_BAR_FORMAT, repr_phi_theta
from ..map import MAP_QUANTITIES, ProjectionMap
from ..tod import TOD, TOD_QUANTITIES
from ..units import Quantity, parse_units

# np.seterr(invalid="ignore")

here, this_filename = os.path.split(__file__)
logger = logging.getLogger("maria")


class BaseMapper:
    """
    The base class for mapping, which handles both projection maps and HEALPix maps.
    """

    def __init__(
        self,
        tods: Sequence[TOD],
        resolution: Quantity,
        units: str,
        stokes: str,
        min_time: float,
        max_time: float,
        timestep: float,
        tod_preprocessing: Mapping,
        map_postprocessing: Mapping,
        progress_bars: bool,
    ):
        u = parse_units(units)
        self.tod_units = units if u["quantity"] in TOD_QUANTITIES else "K_RJ"
        self.map_units = units

        if stokes is None:
            stokes = "IQUV" if any([tod.dets.polarized for tod in tods]) else "I"
            logger.info(f"Inferring mapper stokes parameters '{stokes}' for mapper.")

        if u["quantity"] not in MAP_QUANTITIES:
            raise ValueError(f"Units '{units}' (with associated quantity '{u['quantity']}') are not valid map units")

        self.resolution = resolution
        self.units = units
        self.stokes = stokes
        self.tod_preprocessing = tod_preprocessing
        self.map_postprocessing = map_postprocessing
        self.progress_bars = progress_bars

        min_time = min_time or min([tod.coords.t.min() for tod in tods])
        max_time = max_time or max([tod.coords.t.max() for tod in tods])
        mean_time = (min_time + max_time) / 2

        if timestep is None:
            self.timestep = np.inf
            self.t = np.array([mean_time])
        else:
            self.timestep = timestep
            time_bins = np.arange(min_time, max_time + timestep, timestep)
            time_bins += mean_time - self.time_bins.mean()
            self.t = (time_bins[1:] + time_bins[:-1]) / 2

        self.bands = BandList([])

        self.tods = []
        self.add_tods(tods)

        beam_sum = np.zeros((len(self.bands), 1, 3))
        beam_wgt = np.zeros((len(self.bands), 1, 3))

        for band_index, band in enumerate(self.bands):
            for tod in self.tods:
                beam_sum[band_index] += tod.duration * tod.dets.beams[tod.dets.band_name == band.name].mean(axis=0)
                beam_wgt[band_index] += tod.duration

        self.beam = beam_sum / beam_wgt

    @property
    def nu(self):
        return np.unique([nu.Hz for nu in self.bands.center])

    @property
    def nu_bins(self):
        return np.array([0, *(self.nu[1:] + self.nu[:-1]), np.inf])

    @property
    def t_bins(self):
        return np.array([-np.inf, *(self.t[1:] + self.t[:-1]), np.inf])

    @property
    def n_stokes(self):
        return len(self.stokes)

    @property
    def n_bands(self):
        return len(self.bands)

    @property
    def n_t(self):
        return len(self.t)

    @property
    def map(self):
        return ProjectionMap(
            data=self.data.reshape(self.map_shape),
            weight=self.weight.reshape(self.map_shape),
            stokes=self.stokes,
            t=self.timestamps,
            nu=self.bands.center,
            resolution=self.resolution,
            center=self.center,
            degrees=False,
            frame=self.frame.name,
            units=self.tod_units,
            beam=self.beam,
        ).to(self.map_units)

    def add_tods(self, tods):
        tods_pbar = tqdm(
            np.atleast_1d(tods), desc="Preprocessing TODs", bar_format=DEFAULT_BAR_FORMAT, disable=not self.progress_bars
        )
        for tod in tods_pbar:
            self.tods.append(tod.process(config=self.tod_preprocessing).to(self.tod_units))
            for band in tod.dets.bands:
                self.bands.add(band)

    def initialize_mapper(self):
        raise NotImplementedError()

    def plot(self):
        if not hasattr(self, "map"):
            raise RuntimeError("Mapper has not been run yet.")
        self.map.plot()

    def _run(self):
        """
        A method to be overwritten by the specific mapping procedure.

        It should return a dict with the raw map and weight.
        """
        raise ValueError("Not implemented!")

    def run(self):
        if not len(self.tods):
            raise RuntimeError("This mapper has no TODs!")

        map_data = self._run()
        map_data["sum"] = map_data["data"] * map_data["weight"]

        # if "median_filter" in self.map_postprocessing.keys():
        #     size = self.map_postprocessing["median_filter"]["size"]
        #     map_data["data"] = sp.ndimage.median_filter(
        #         map_data["data"],
        #         size=size,
        #         axes=(-2, -1),
        #     )

        if "gaussian_filter" in self.map_postprocessing.keys():
            sigma = self.map_postprocessing["gaussian_filter"]["sigma"]

            map_data["sum"] = sp.ndimage.gaussian_filter(map_data["sum"], sigma=(0, 0, 0, sigma, sigma))
            map_data["weight"] = sp.ndimage.gaussian_filter(map_data["weight"], sigma=(0, 0, 0, sigma, sigma))

        self.data = (map_data["sum"] / map_data["weight"]).reshape(self.map_shape)
        self.weight = map_data["weight"].reshape(self.map_shape)

        for stokes_index, stokes in enumerate(self.stokes):
            for nu_index, nu in enumerate(self.nu):
                if map_data["weight"][stokes_index, nu_index].sum() == 0:
                    logger.warning(f"No counts for map (stokes={stokes}, nu={Quantity(nu, 'Hz')})")

        # by convention maps have zero mean
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            self.data -= np.nanmean(self.data, axis=(-1, -2))[..., None, None]

        return self.map

    @property
    def map(self):
        raise NotImplementedError()


class BaseProjectionMapper(BaseMapper):
    """
    The base class for making a Projection map.
    """

    def __init__(
        self,
        tods: Sequence[TOD],
        center: tuple[Quantity, Quantity],
        stokes: str,
        width: Quantity,
        height: Quantity,
        resolution: Quantity,
        frame: str,
        units: str,
        tod_preprocessing: Mapping,
        map_postprocessing: Mapping,
        min_time: float,
        max_time: float,
        timestep: float,
        degrees: bool,
        progress_bars: bool,
    ):
        center = (Quantity(center, "deg" if degrees else "rad")) if center is not None else None
        width = (Quantity(width, "deg" if degrees else "rad")) if width is not None else None
        height = (Quantity(height, "deg" if degrees else "rad")) if height is not None else None
        resolution = (Quantity(resolution, "deg" if degrees else "rad")) if resolution is not None else None

        infer_center, infer_width, infer_height = infer_center_width_height(
            coords_list=[tod.coords for tod in tods], center=center, frame=frame, square=True
        )

        logger.debug(
            f"Inferred center={Quantity(infer_center, 'rad')}, width={Quantity(infer_width, 'rad')}, \
                     width={Quantity(infer_height, 'rad')} for map."
        )

        if center is None:
            center = Quantity(infer_center, "rad")
            logger.info(
                f"Inferring center {repr_phi_theta(phi=center[0].rad, theta=center[1].rad, frame=frame)} for mapper."
            )

        if width is None:
            if height is not None:
                width = height
                logger.info(f"Inferring mapper width {width} to match supplied height.")
            else:
                width = Quantity(infer_width, "rad")
                logger.info(f"Inferring mapper width {width} for mapper from observation patch.")

        if height is None:
            if width is not None:
                height = width
                logger.info(f"Inferring mapper height {height} to match supplied width.")
            else:
                height = Quantity(infer_height, "rad")
                logger.info(f"Inferring mapper height {height} for mapper from observation patch.")

        if resolution is None:
            resolution = Quantity(width / 128, "rad")
            logger.info(f"Inferring mapper resolution {resolution} for mapper from observation patch.")

        super().__init__(
            tods=tods,
            resolution=resolution,
            units=units,
            stokes=stokes,
            min_time=min_time,
            max_time=max_time,
            timestep=timestep,
            tod_preprocessing=tod_preprocessing,
            map_postprocessing=map_postprocessing,
            progress_bars=progress_bars,
        )

        self.center = center
        self.width = width
        self.height = height
        self.frame = Frame(frame)

        self.data = np.nan * np.zeros(self.map_shape)
        self.weight = np.nan * np.ones(self.map_shape)

    @property
    def n_x(self):
        return int(np.maximum(1, self.width / self.resolution))

    @property
    def n_y(self):
        return int(np.maximum(1, self.height / self.resolution))

    @property
    def x_bins(self):
        return np.linspace(-0.5 * self.width.rad, 0.5 * self.width.rad, self.n_x + 1)

    @property
    def y_bins(self):
        return np.linspace(0.5 * self.height.rad, -0.5 * self.height.rad, self.n_y + 1)

    @property
    def map_shape(self):
        return (self.n_stokes, len(self.nu), self.n_t, self.n_y, self.n_x)

    @property
    def map_size(self):
        return np.prod(self.map_shape)

    @property
    def map(self):
        return ProjectionMap(
            data=self.data,
            weight=self.weight,
            stokes=self.stokes,
            t=self.t,
            nu=self.nu,
            resolution=self.resolution,
            center=self.center,
            degrees=False,
            frame=self.frame.name,
            units=self.tod_units,
            beam=self.beam,
        ).to(self.map_units)



from __future__ import annotations

import copy
import os
from typing import Mapping

import numpy as np
import pandas as pd

from ..band import BandList  # noqa
from ..beam import compute_angular_fwhm
from ..units import Quantity
from ..utils import HEX_CODE_LIST, compute_diameter
from .array import Array, get_array, get_array_config

here, this_filename = os.path.split(__file__)


class ArrayList:
    def __init__(self, arrays: list[Mapping]):
        if isinstance(arrays, Mapping):
            # convert to a list
            arrays = [{"name": k, **v} for k, v in arrays.items()]

        if isinstance(arrays, list):
            unnamed_array_index = 1
            self.arrays = []
            for array in arrays:
                if isinstance(array, str):
                    self.arrays.append(get_array(array))
                elif isinstance(array, Mapping):
                    if "name" not in array:
                        array["name"] = f"array{unnamed_array_index}"
                        unnamed_array_index += 1
                    array_config = get_array_config(**array)
                    self.arrays.append(Array.from_config(array_config))
                elif isinstance(array, Array):
                    self.arrays.append(array)
                else:
                    raise ValueError("Arrays must be either a string or a mapping")
        elif isinstance(arrays, ArrayList):
            self.arrays = arrays.arrays
        else:
            raise ValueError("Each element of 'arrays' must be either an Array, a string, or a mapping.")

    def combine(self):
        array_dets = []
        for array in self.arrays:
            df = copy.deepcopy(array.dets)
            df.loc[:, "array_name"] = array.name
            array_dets.append(df)
        return Array(name="", dets=pd.concat(array_dets), bands=self.bands)

    def one_detector_from_each_band(self):
        return ArrayList(arrays=[array.one_detector_from_each_band() for array in self.arrays])

    def outer(self):
        return ArrayList(arrays=[array.outer() for array in self.arrays])

    @property
    def field_of_view(self):
        return Quantity(compute_diameter(self.offsets), units="rad")

    @property
    def max_baseline(self):
        return Quantity(compute_diameter(self.baselines), units="m")

    @property
    def n(self):
        return sum([array.n for array in self.arrays])

    @property
    def dets(self):
        return pd.concat([array.dets for array in self.arrays])

    @property
    def bands(self):
        bands = []
        for array in self.arrays:
            for band in array.bands:
                if band not in bands:
                    bands.append(band)
        return BandList(bands)

    def angular_fwhm(self, z):  # noqa F401
        """
        Angular beam width (in radians) as a function of depth (in meters)
        """
        return Quantity(compute_angular_fwhm(z=z, fwhm_0=self.primary_size, n=1, nu=self.band_center), "rad")

    def physical_fwhm(self, z):
        """
        Physical beam width (in meters) as a function of depth (in meters)
        """
        return Quantity(z * self.angular_fwhm(z).rad)

    def mask(self, **kwargs):
        return np.concatenate([array.mask(**kwargs) for array in self.arrays], axis=0)

    def subset(self, **kwargs):
        return ArrayList([array.subset(**kwargs).dets for array in self.arrays])

    def summary(self):
        return pd.concat([array.summary() for array in self.arrays], axis=1).T

    @property
    def array_name(self):
        return np.concatenate([array.n * [array.name] for array in self.arrays], axis=0)

    @property
    def offsets(self):
        return np.concatenate([array.offsets for array in self.arrays], axis=0)

    @property
    def baselines(self):
        return np.concatenate([array.baselines for array in self.arrays], axis=0)

    def passband(self, nu):
        return np.concatenate([array.passband(nu) for array in self.arrays], axis=0)

    def __getitem__(self, key):
        return self.arrays[key]

    def __getattr__(self, attr):
        try:
            return np.concatenate([getattr(array, attr) for array in self.arrays], axis=0)
        except Exception:
            pass
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{attr}'")

    def __repr__(self):
        return self.summary().__repr__()

    def _repr_html_(self):
        return self.summary()._repr_html_()

    def __iter__(self):  # it has to be called this
        return iter(self.arrays)  # return the list's iterator

    def __len__(self):
        return len(self.arrays)



import os

import dask.array as da
import h5py
import healpy as hp
import matplotlib.pyplot as plt
import numpy as np
import scipy as sp

from ..coords import FRAMES, Coordinates, Frame
from ..units import Quantity
from .base import Map

here, this_filename = os.path.split(__file__)


class HEALPixMap(Map):
    """
    A spherical map. It has shape (stokes, nu, t, pix).
    """

    def __init__(
        self,
        data: float,
        weight: float = None,
        stokes: float = None,
        nu: float = None,
        t: float = None,
        z: float = None,
        beam: tuple[float, float, float] = None,
        frame: str = "ra/dec",
        units: str = "K_RJ",
        degrees: bool = True,
        dtype: type = np.float32,
    ):
        if weight is not None:
            if weight.shape != data.shape:
                raise ValueError(f"'data' {data.shape} and 'weight' {weight.shape} should have the same shape.")
        else:
            weight = da.ones_like(data)

        map_dims = {"npix": data.shape[-1]}

        beam = beam if beam is not None else 0.0

        super().__init__(
            data=data,
            weight=weight,
            stokes=stokes,
            nu=nu,
            t=t,
            z=z,
            beam=beam,
            map_dims=map_dims,
            units=units,
            degrees=degrees,
            dtype=dtype,
        )

        if not hp.pixelfunc.isnpixok(self.npix):
            raise ValueError(f"Invalid pixel count (n={self.npix}).")

        self.nside = hp.pixelfunc.npix2nside(self.npix)
        self.frame = Frame(frame)

        if not hasattr(beam, "__len__"):
            beam = beam or self.resolution
            beam = (beam, beam, 0)

        if len(beam) != 3:
            raise ValueError("'beam' must be either a number or a tuple of (major, minor, angle)")

        self.beam = tuple(np.radians(beam) if degrees else beam)

    def pointing_matrix(self, coords: Coordinates):
        pixel_index = hp.ang2pix(
            nside=self.nside,
            phi=getattr(coords, self.frame.phi_name).ravel(),
            theta=np.pi / 2 - getattr(coords, self.frame.theta_name).ravel(),
        ).ravel()

        return sp.sparse.csr_array(
            (np.ones(coords.size, dtype=np.uint8), (np.arange(coords.size), pixel_index)), shape=(coords.size, self.npix)
        )

    @property
    def resolution(self):
        return hp.pixelfunc.nside2resol(self.nside)

    @property
    def pixel_area(self):
        return Quantity(hp.pixelfunc.nside2resol(self.nside) ** 2, "sr")

    @property
    def npix(self):
        return self.dims["npix"]

    def package(self):
        return {
            "data": self.data,
            "weight": self.weight,
            "stokes": self.stokes,
            "nu": self.nu,
            "t": self.t,
            "frame": self.frame.name,
            "units": self.units,
        }

    @property
    def X(self):
        return np.meshgrid(self.x_side, self.y_side)[0]

    @property
    def Y(self):
        return np.meshgrid(self.x_side, self.y_side)[1]

    def smooth(self, sigma: float = None, fwhm: float = None, inplace: bool = False):
        if not (sigma is None) ^ (fwhm is None):
            raise ValueError("You must supply exactly one of 'sigma' or 'fwhm'.")

        sigma = sigma if sigma is not None else fwhm / np.sqrt(8 * np.log(2))

        data = np.stack([hp.sphtfunc.smoothing(m, sigma=sigma) for m in self.data.reshape(-1, self.npix)]).reshape(
            self.data.shape
        )

        if inplace:
            self.data = data

        else:
            package = self.package()
            package.update({"data": data})
            return type(self)(**package)

    def to_hdf(self, filename):
        with h5py.File(filename, "w") as f:
            f.create_dataset("data", dtype=float, data=self.data)

            if self._weight is not None:
                f.create_dataset("weight", dtype=float, data=self._weight)

            f.create_dataset("frame", data=self.frame.name)

            for field in ["stokes", "nu", "t", "units"]:
                f.create_dataset(field, data=getattr(self, field))

    def plot(self):
        for i in range(len(self.stokes)):
            m = self.data[i, 0, 0]
            min, max = da.percentile(m, q=[0.1, 99.9])

            hp.newvisufunc.projview(
                m.compute(),
                min=min,
                max=max,
                cmap="cmb",
                sub=(len(self.stokes), 1, 1 + i),
            )
            # hp.visufunc.gnomview(self.data[i,0,0], cmap="cmb")

        plt.tight_layout()

    def __repr__(self):
        return f"""{self.__class__.__name__}:
  shape{self.dims_string}: {self.data.shape}
  stokes: {self.stokes if "stokes" in self.dims else "naive"}
  nu: {self.nu if "nu" in self.dims else "naive"}
  t: {self.t if "t" in self.dims else "naive"}
  z: {self.z if "z" in self.dims else "naive"}
  nside: {self.nside}
  frame: {self.frame.name}
  quantity: {self.u["quantity"]}
  units: {self.units}
    min: {np.nanmin(self.data).compute():.03e}
    max: {np.nanmax(self.data).compute():.03e}
  resolution: {Quantity(self.resolution, "rad")}
  memory: {Quantity(self.data.nbytes + self.weight.nbytes, "B")}"""



from __future__ import annotations

import healpy as hp
import numpy as np

from ..io import download_from_url
from ..map import HEALPixMap

CMB_SPECTRUM_SOURCE_URL = (
    "https://github.com/thomaswmorris/maria-data/raw/master/cmb/spectra/"
    "COM_PowerSpect_CMB-base-plikHM-TTTEEE-lowl-lowE-lensing-minimum-theory_R3.01.txt"
)
CMB_SPECTRUM_CACHE_PATH = "/tmp/maria-data/cmb/spectrum.txt"
CMB_SPECTRUM_CACHE_MAX_AGE = 30 * 86400  # one month

CMB_MAP_SOURCE_URL = (
    "https://irsa.ipac.caltech.edu/data/Planck/release_3/all-sky-maps/maps/component-maps/cmb/"
    "COM_CMB_IQU-143-fgsub-sevem_2048_R3.00_full.fits"
)
CMB_MAP_CACHE_PATH = "/tmp/maria-data/cmb/planck/map.fits"
CMB_MAP_CACHE_MAX_AGE = 30 * 86400  # one month


CMB_SOURCES = {
    "planck": {"spectrum": "cmb/spectra/planck.csv"},
    "camb": {"spectrum": "/Users/tom/maria/data/cmb/spectra/camb.csv"},
}


class CMB(HEALPixMap):
    def __init__(
        self,
        data: float,
        weight: float = None,
        stokes: float = None,
        frame: str = "galactic",
        nu: float = None,
        units: str = "K_CMB",
    ):
        super().__init__(data=data, weight=weight, stokes=stokes, nu=nu, z=1100.0, units=units, frame=frame)


def get_cmb(**kwargs):
    download_from_url(
        source_url=CMB_MAP_SOURCE_URL,
        cache_path=CMB_MAP_CACHE_PATH,
    )

    field_dtypes = {
        "T": np.float32,
        "Q": np.float32,
        "U": np.float32,
        "T_mask": bool,
        "P_mask": bool,
    }

    maps = {
        field: hp.fitsfunc.read_map(CMB_MAP_CACHE_PATH, field=i).astype(dtype)
        for i, (field, dtype) in enumerate(field_dtypes.items())
    }

    maps["T"] = np.where(maps["T_MASK"], maps["T"], np.nan)
    maps["Q"] = np.where(maps["P_MASK"], maps["Q"], np.nan)
    maps["U"] = np.where(maps["P_MASK"], maps["U"], np.nan)

    return CMB(data=[maps["T"], maps["Q"], maps["U"]], fields=["T", "Q", "U"])



from __future__ import annotations

import os
from pathlib import Path

import pandas as pd
from astropy.coordinates import EarthLocation

from ..utils import read_yaml
from .site import REGIONS, InvalidRegionError, Site, all_regions  # noqa

here, this_filename = os.path.split(__file__)

SITE_CONFIGS = {}
for sites_path in Path(f"{here}/sites").glob("*.yml"):
    SITE_CONFIGS.update(read_yaml(sites_path))

for site in SITE_CONFIGS:
    config = SITE_CONFIGS[site]
    SITE_CONFIGS[site]["instruments"] = ", ".join(config["instruments"]) if "instruments" in config else ""

SITE_DISPLAY_COLUMNS = [
    "description",
    "instruments",
    "region",
    "latitude",
    "longitude",
    "altitude",
]
site_data = pd.DataFrame(SITE_CONFIGS).T.sort_values("region")
all_sites = list(site_data.index.values)


class InvalidSiteError(Exception):
    def __init__(self, invalid_site):
        super().__init__(
            f"The site '{invalid_site}' is not supported. "
            f"Supported sites are:\n\n{site_data.loc[:, SITE_DISPLAY_COLUMNS].to_string()}",
        )


def get_location(site_name):
    site = get_site(site_name)
    return EarthLocation.from_geodetic(
        lon=site.longitude.deg,
        lat=site.latitude.deg,
        height=site.altitude.m,
    )


def get_site_config(site_name="hoagie_haven", **kwargs):
    if site_name not in SITE_CONFIGS.keys():
        raise InvalidSiteError(site_name)
    SITE_CONFIG = SITE_CONFIGS[site_name].copy()
    for k, v in kwargs.items():
        SITE_CONFIG[k] = v
    return SITE_CONFIG


def get_site(site_name="hoagie_haven", **kwargs):
    return Site(**get_site_config(site_name=site_name, **kwargs))



import copy
import logging
import os
from typing import Callable, Iterable

import astropy as ap
import dask.array as da
import h5py
import jax
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import scipy as sp
from astropy.io import fits
from astropy.wcs import WCS
from skimage.measure import block_reduce

from ..array import Array
from ..coords import FRAMES, Coordinates, Frame
from ..io import repr_phi_theta
from ..tod import TOD
from ..units import Quantity, get_factor_and_base_units_vector
from ..utils import compute_pointing_matrix_sparse_indices, unpack_implicit_slice
from .base import Map

here, this_filename = os.path.split(__file__)
logger = logging.getLogger("maria")


class ProjectionMap(Map):
    """
    A rectangular map projected on the sphere. It has shape (stokes, nu, t, y, x).
    """

    def __init__(
        self,
        data: float,
        weight: float = None,
        stokes: float = None,
        nu: Iterable[float] = None,
        t: Iterable[float] = None,
        z: Iterable[float] = None,
        width: float = None,
        height: float = None,
        resolution: float = None,
        x_res: float = None,
        y_res: float = None,
        center: tuple[float, float] = (0.0, 0.0),
        beam: tuple[float, float, float] = 0.0,
        frame: str = "ra/dec",
        degrees: bool = True,
        units: str = "K_RJ",
        dtype: type = np.float32,
    ):
        # give it five dimensions

        if data.ndim < 2:
            raise ValueError("'data' must be at least two-dimensional")

        if weight is not None:
            if weight.shape != data.shape:
                raise ValueError(f"'data' {data.shape} and 'weight' {weight.shape} should have the same shape.")
        else:
            weight = da.ones_like(data)

        map_dims = {"y": data.shape[-2], "x": data.shape[-1]}

        try:
            assert len(center) == 2
            self.center = tuple([*Quantity(center, ("deg" if degrees else "rad")).pin("deg")])
        except Exception:
            raise ValueError("'center' must be a two-tuple of numbers")

        self.frame = Frame(frame)
        beam = beam if beam is not None else 0.0

        super().__init__(
            data=data,
            weight=weight,
            stokes=stokes,
            nu=nu,
            t=t,
            z=z,
            beam=beam,
            map_dims=map_dims,
            units=units,
            degrees=degrees,
            dtype=dtype,
        )

        if x_res is not None:
            y_res = y_res or x_res
        elif y_res is not None:
            x_res = x_res or y_res
        elif resolution is not None:
            if not resolution > 0:
                raise ValueError("'resolution' must be positive.")
            x_res = y_res = resolution
        else:
            if width is not None:
                if not width > 0:
                    raise ValueError("'width' must be positive.")
                x_res = width / self.n_x
                if height is not None:
                    y_res = height / self.n_y
                else:
                    y_res = x_res
            else:
                # here height must not be None
                x_res = y_res = height / self.n_y

        self.x_res = Quantity(x_res, "deg" if degrees else "rad")
        self.y_res = Quantity(y_res, "deg" if degrees else "rad")

        if self.x_res.deg < 0:
            self.data, self.weight = self.data[::-1], self.weight[::-1]
            self.x_res *= -1
        if self.y_res.deg < 0:
            self.data, self.weight = self.data[::-1, :], self.weight[::-1, :]
            self.y_res *= -1

        if all(x is None for x in [width, height, resolution, x_res, y_res]):
            raise ValueError("You must pass at least one of 'width', 'height', 'resolution', 'x_res', or 'y_res'.")

    def _pointing_matrix_sparse_indices(self, coords: Coordinates):
        offsets = coords.offsets(center=(self.center[0].rad, self.center[1].rad), frame=self.frame.name)

        if "t" in self.dims:
            return compute_pointing_matrix_sparse_indices(
                x_list=(offsets[..., 0], offsets[..., 1], coords._t), bins_list=(self.x_bins, self.y_bins[::-1], self.t_bins)
            )
        else:
            return compute_pointing_matrix_sparse_indices(
                x_list=(offsets[..., 0], offsets[..., 1]), bins_list=(self.x_bins, self.y_bins[::-1])
            )

    def compute_pointing_matrix_sparse_indices(self, coords: Coordinates, dets: Array):
        nd, nt = coords.shape
        offsets = coords.offsets(frame=self.frame, center=self.center)

        dim_index = {}

        dim_index["x"] = np.digitize(offsets[..., 0], self.x_bins).ravel() - 1
        dim_index["y"] = np.digitize(offsets[..., 1], self.y_bins[::-1]).ravel() - 1
        dim_index["t"] = np.digitize(coords._t, self.t_bins).ravel() - 1
        dim_index["nu"] = np.digitize(np.repeat(dets.band_center, nt), self.nu_bins) - 1

        cum_dim = 1
        map_index = 0
        mask = True
        for dim, index in dim_index.items():
            n = self.dims[dim]
            mask &= (index > -1) & (index < n)
            map_index += cum_dim * index
            cum_dim *= n
        pixel_index = map_index[mask]
        sample_index = np.arange(coords.size)[mask]

        values_list = []
        pixel_index_list = []
        sample_index_list = []
        M = dets.mueller()
        for stokes_index, stokes in enumerate(self.stokes):
            values_list.append(np.repeat(M[:, 0, "IQUV".index(stokes)], nt)[mask])
            pixel_index_list.append(pixel_index + cum_dim * stokes_index)
            sample_index_list.append(sample_index)
        return np.concatenate(values_list), np.concatenate(pixel_index_list), np.concatenate(sample_index_list)

    def pointing_matrix(self, coords: Coordinates, dets: Array):
        values, pixel_index, sample_index = self.compute_pointing_matrix_sparse_indices(coords=coords, dets=dets)
        return sp.sparse.csr_array((values, (sample_index, pixel_index)), shape=(coords.size, self.data.size))

    def header(self):
        header = ap.io.fits.header.Header()

        header["SIMPLE"] = "T / conforms to FITS standard"
        header["BITPIX"] = "-32 / array data type"
        header["NAXIS"] = len(self.dims)
        for dim_index, n in enumerate(list(self.dims.values())):
            header[f"NAXIS{dim_index + 1}"] = n

        if "nu" in self.dims:
            header["RESTFREQ"] = self.nu.Hz.mean()

        header["CDELT1"] = -1 * self.x_res.degrees  # degrees
        header["CDELT2"] = self.y_res.degrees  # degrees
        header["CRPIX1"] = self.data.shape[-1] // 2
        header["CRPIX2"] = self.data.shape[-2] // 2
        header["BUNITS"] = self.units
        header["BMAJOR"] = self.beam[..., 0].degrees.item()
        header["BMINOR"] = self.beam[..., 1].degrees.item()
        header["BPA"] = self.beam[..., 2].degrees.item()

        # specify x center
        CTYPE1 = self.frame.fits_phi
        header["CTYPE1"] = f"{CTYPE1}{(5 - len(CTYPE1)) * '-'}SIN"
        header["CRVAL1"] = self.center[0].deg
        header["CUNIT1"] = "deg     "

        # center y center
        CTYPE2 = self.frame.fits_theta
        header["CTYPE2"] = f"{CTYPE2}{(5 - len(CTYPE2)) * '-'}SIN"
        header["CRVAL2"] = self.center[1].deg
        header["CUNIT2"] = "deg     "

        return header

    def __getattr__(self, attr):
        broadcasted_attrs = ["STOKES", "NU", "T", "Y", "X"]
        if attr in broadcasted_attrs:
            broadcasted_attr_values = np.meshgrid(self.stokes, self.nu.Hz, self.t, self.y_side, self.x_side)
            return broadcasted_attr_values[broadcasted_attrs.index(attr)]
        if attr in ["min", "max", "rms"]:
            if not hasattr(self, "stats"):
                self.compute_stats()
            return self.stats[attr]

        raise AttributeError(f"'ProjectionMap' object has no attribute '{attr}'")

    def __getitem__(self, key):
        if not hasattr(key, "__iter__"):
            key = (key,)

        explicit_slices = unpack_implicit_slice(key, ndims=self.ndim)

        _, dimensions = get_factor_and_base_units_vector(self.units)
        if all([dimensions[dim] == 0 for dim in ["pixel", "rad"]]):
            # don't convert if in resolution-neutral units (like temperature)
            package = self.package()
        else:
            # convert if in non-resolution-neutral units (like spectral flux)
            package = self.to("K_RJ").package()

        package["data"] = package["data"][key]
        package["weight"] = package["weight"][key]

        for axis, (dim, naxis) in enumerate(self.dims.items()):
            if dim not in "xy":
                package[dim] = package[dim][explicit_slices[axis]]

            if dim in "nu":
                package["beam"] = package["beam"][explicit_slices[axis]]

        package["y_res"] *= explicit_slices[-2].step or 1
        package["x_res"] *= explicit_slices[-1].step or 1

        return ProjectionMap(**package).to(self.units)

    def downsample(self, reduce: tuple, func: Callable = np.mean):
        if reduce:
            if len(reduce) != 4:
                raise ValueError("'reduce' must be a four-tuple of ints (nu, t, y, x)")

            package = self.to("K_RJ").package()
            package["data"] = block_reduce(package["data"].compute(), block_size=(1, *reduce), func=func)
            package["weight"] = block_reduce(package["weight"].compute(), block_size=(1, *reduce), func=func)

            *_, new_n_y, new_n_x = package["data"].shape

            package["nu"] = block_reduce(package["nu"], block_size=reduce[0])
            package["t"] = block_reduce(package["t"], block_size=reduce[1])
            package["x_res"] *= self.n_x / new_n_x
            package["y_res"] *= self.n_y / new_n_y

            return ProjectionMap(**package).to(self.units)

    @property
    def points(self):
        return np.stack(np.meshgrid(self.y_side, self.x_side, indexing="ij"), axis=-1)

    def compute_stats(self):
        self.stats = {
            "min": np.nanmin(self.data).compute(),
            "max": np.nanmax(self.data).compute(),
            "rms": np.sqrt(np.sum(np.square(self.data - self.data.mean()) * self.weight / self.weight.sum())).compute(),
        }

    def __repr__(self):
        # beam_repr = self.beam
        center_repr = "center:"
        for key, value in repr_phi_theta(self.center[0].rad, self.center[1].rad, frame=self.frame.name).items():
            center_repr += f"\n    {key}: {value}"
        return f"""{self.__class__.__name__}:
  shape{self.dims_string}: {self.data.shape}
  stokes: {self.stokes if "stokes" in self.dims else "naive"}
  nu: {self.nu if "nu" in self.dims else "naive"}
  t: {self.t if "t" in self.dims else "naive"}
  z: {self.z if "z" in self.dims else "naive"}
  quantity: {self.u["quantity"]}
  units: {self.units}
    min: {self.min:.03e}
    max: {self.max:.03e}
    rms: {self.rms:.03e}
  {center_repr}
  size(y, x): ({self.height}, {self.width})
  resolution(y, x): ({self.y_res}, {self.x_res})
  beam(maj, min, rot): {self.beam}
  memory: {Quantity(self.data.nbytes + self.weight.nbytes, "B")}"""

    def package(self):
        package = copy.deepcopy(
            {
                "degrees": True,
                "data": self.data,
                "weight": self.weight,
                "stokes": self.stokes,
                "nu": self.nu.Hz,
                "t": self.t,
                "center": (self.center[0].deg, self.center[1].deg),
                "frame": self.frame.name,
                "units": self.units,
                "x_res": self.x_res.deg,
                "y_res": self.y_res.deg,
                "beam": self.beam.deg,
            }
        )

        for dim in ["stokes", "nu", "t"]:
            if dim not in self.dims:
                package.pop(dim)

        return package

    @property
    def pixel_area(self):
        return Quantity(self.x_res.rad * self.y_res.rad, "sr")

    @property
    def resolution(self):
        if not np.isclose(self.x_res.rad, self.y_res.rad, rtol=1e-3):
            RuntimeError(
                "Cannot return attribute 'resolution'; ProjectionMap has x-resolution"
                f" {self.x_res}° and y-resolution {self.x_res}°."
            )
        return self.x_res

    @property
    def width(self):
        return self.x_res * self.n_x

    @property
    def height(self):
        return self.y_res * self.n_y

    @property
    def n_y(self):
        return self.data.shape[-2]

    @property
    def n_x(self):
        return self.data.shape[-1]

    @property
    def x_bins(self):
        return self.width.rad * np.linspace(-0.5, 0.5, self.n_x + 1)

    @property
    def y_bins(self):
        """
        The negative is so that follows indexing rules
        """
        return -self.height.rad * np.linspace(-0.5, 0.5, self.n_y + 1)

    @property
    def x_side(self):
        return (self.x_bins[:-1] + self.x_bins[1:]) / 2

    @property
    def y_side(self):
        return (self.y_bins[:-1] + self.y_bins[1:]) / 2

    def smooth(self, sigma: float = None, fwhm: float = None):
        if not (sigma is None) ^ (fwhm is None):
            raise ValueError("You must supply exactly one of 'sigma' or 'fwhm'.")

        package = self.package()

        sigma = sigma if sigma is not None else fwhm / np.sqrt(8 * np.log(2))
        x_sigma_pixels = sigma / self.x_res
        y_sigma_pixels = sigma / self.y_res
        package["data"] = sp.ndimage.gaussian_filter(self.data, sigma=(y_sigma_pixels, x_sigma_pixels), axes=(-2, -1))

        return type(self)(**package)

    def plot_slice(
        self,
        fig,
        ax,
        nu_index: int = 0,
        t_index: int = 0,
        stokes: str = "I",
        cmap: str = "default",
        rel_vmin: float = 0.005,
        rel_vmax: float = 0.995,
        norm_method: str = "slice",
        vmin: float = None,
        vmax: float = None,
    ):
        d = self.data
        w = self.weight

        logger.debug(f"Plotting map slice (stokes={stokes}, nu_index={nu_index}, t_index={t_index})")

        if "stokes" in self.dims:
            if stokes not in self.stokes:
                raise ValueError(f"Invalid stokes parameter '{stokes}'. This map has stokes parameters '{self.stokes}'.")
            stokes_index = self.stokes.index(stokes)
            d, w = d[stokes_index], w[stokes_index]

        if "nu" in self.dims:
            try:
                d, w = d[nu_index], w[nu_index]
            except IndexError:
                raise IndexError(
                    f"nu_index={nu_index} is out of bounds; map has shape {self.dims_string} = {tuple(self.dims.values())}"
                )

        if "t" in self.dims:
            try:
                d, w = d[t_index], w[t_index]
            except IndexError:
                raise IndexError(
                    f"t_index={t_index} is out of bounds; map has shape {self.dims_string} = {tuple(self.dims.values())}"
                )

        if cmap == "default":
            cmap = "CMRmap" if stokes == "I" else "cmb"

        map_slice_qdata = Quantity(d.compute(), units=self.units)

        if vmin is None or vmax is None:
            if norm_method == "slice":
                subset = np.random.choice(d.size, size=min(d.size, 20000), replace=False)
                auto_vmin, auto_vmax = np.nanquantile(
                    map_slice_qdata.value.ravel()[subset],
                    weights=w.ravel()[subset].compute(),
                    q=(rel_vmin, rel_vmax),
                    method="inverted_cdf",
                )
            elif norm_method == "total":
                map_values = Quantity(self.data.compute(), units=self.units).to(map_slice_qdata.units)
                subset = np.random.choice(map_values.size, size=min(d.size, 20000), replace=False)
                auto_vmin, auto_vmax = np.nanquantile(
                    map_values.ravel()[subset],
                    weights=self.weight.ravel()[subset].compute(),
                    q=(rel_vmin, rel_vmax),
                    method="inverted_cdf",
                )
            else:
                raise ValueError(f"Invalid normalization method '{norm_method}'")

            vmin = vmin or auto_vmin
            vmax = vmax or auto_vmax

        X = np.r_[self.x_bins, self.y_bins]
        grid_u = Quantity(X, "rad").u

        x = Quantity(self.x_bins, "rad")
        y = Quantity(self.y_bins, "rad")

        # fig = plt.figure(figsize=(6, 6), dpi=256, constrained_layout=True)
        # ax = fig.add_subplot(nx, ny, index, projection=WCS(header))

        ax.pcolormesh(
            getattr(x, grid_u["units"]),
            getattr(y, grid_u["units"]),
            map_slice_qdata.value,
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
        )

        ax.grid(color="white", ls="solid", lw=5e-1)

        dummy_map = mpl.cm.ScalarMappable(
            mpl.colors.Normalize(vmin=vmin, vmax=vmax),
            cmap=cmap,
        )

        cbar = fig.colorbar(
            dummy_map,
            ax=ax,
            shrink=0.75,
            aspect=16,
            location="right",
            pad=0,
        )

        label = rf"${map_slice_qdata.u['math_name']}$"
        if "stokes" in self.dims:
            label += f" Stokes {stokes}"
        if "nu" in self.dims:
            label += f" at {self.nu[nu_index]}"
        cbar.set_label(label, fontsize=10)
        ax.tick_params(axis="x", bottom=True, top=False)
        ax.tick_params(axis="y", left=True, right=False, rotation=90)

        lon = ax.coords[0]
        lon.set_axislabel(rf"{self.frame.phi_long_name}")
        lon.set_ticks_position("b")
        lon.set_ticklabel_position("b")
        lon.set_axislabel_position("b")

        lat = ax.coords[1]
        lat.set_axislabel(rf"{self.frame.theta_long_name}")
        lat.set_ticks_position("l")
        lat.set_ticklabel_position("l")
        lat.set_axislabel_position("l")

        ax.set_aspect("equal")

        return ax

    def plot(
        self,
        stokes: str = "I",
        nu_index: int = 0,
        t_index: int = 0,
        cmap: str = "default",
        rel_vmin: float = 0.005,
        rel_vmax: float = 0.995,
        filename: str = None,
        norm_method: str = "slice",
        vmin: float = None,
        vmax: float = None,
    ):
        X = np.r_[self.x_bins, self.y_bins]
        grid_u = Quantity(X, "rad").u

        header = fits.header.Header()
        header["CDELT1"] = -np.degrees(grid_u["factor"])
        header["CDELT2"] = np.degrees(grid_u["factor"])
        header["CRPIX1"] = 1
        header["CRPIX2"] = 1
        header["CTYPE1"] = "RA---SIN"
        header["CUNIT1"] = "deg     "
        header["CTYPE2"] = "DEC--SIN"
        header["CUNIT2"] = "deg     "
        header["RADESYS"] = "FK5     "
        header["CRVAL1"] = self.center[0].deg
        header["CRVAL2"] = self.center[1].deg

        # nu_index = 0
        # t_index = 0
        # stokes = [["I", "Q"],
        #         ["U", "V"]]
        # stokes = ["I", "Q", "U"]
        # stokes = [["I", "Q"]]
        NU_INDEX, T_INDEX, STOKES = [np.atleast_2d(x) for x in np.broadcast_arrays(nu_index, t_index, stokes)]

        nrows, ncols = STOKES.shape

        max_fig_size = 12

        ax_size = 5  # np.minimum(np.maximum(np.minimum(max_fig_size / nrows, max_fig_size / ncols), 5), 8)

        fig, axes = plt.subplots(
            nrows,
            ncols,
            figsize=(ax_size * ncols, ax_size * nrows),
            subplot_kw={"projection": WCS(header)},
            constrained_layout=True,
            sharex=True,
            sharey=True,
        )
        axes = np.atleast_1d(axes).reshape(STOKES.shape)

        for r in range(nrows):
            for c in range(ncols):
                ax = axes[r, c]

                if c > 0:
                    ax.coords[1].set_ticks_visible(False)
                    ax.coords[1].set_ticklabel_visible(False)
                # if c == ncols - 1:
                #     ay2 = ax.secondary_yaxis("right")
                #     ay2.set_ylabel(rf"$\Delta\,\theta_y$ [{grid_u['units']}]")
                if r == 0:
                    ax2 = ax.secondary_xaxis("top")
                    ax2.set_xlabel(rf"$\Delta\,\theta$ [{grid_u['units']}]")
                if r < nrows - 1:
                    ax.coords[0].set_ticks_visible(False)
                    ax.coords[0].set_ticklabel_visible(False)

                if STOKES[r, c] is None:
                    continue

                self.plot_slice(
                    fig,
                    ax,
                    stokes=STOKES[r, c],
                    nu_index=NU_INDEX[r, c],
                    t_index=T_INDEX[r, c],
                    rel_vmin=rel_vmin,
                    rel_vmax=rel_vmax,
                    cmap=cmap,
                    norm_method=norm_method,
                    vmin=vmin,
                    vmax=vmax,
                )

                if filename is not None:
                    plt.savefig(fname=filename, dpi=256)

    def to_fits(self, filename):
        if self.dims.get("nu", np.nan) > 1:
            raise RuntimeError("Cannot write multifrequency maps to FITS")

        m = self  # .to(self.u["base_unit"])

        data = m.data[::-1, :]  # FITS counts from the bottom, while normal people count from the top
        if self.frame.name in ["ra/dec", "galatic"]:
            data = data[..., ::-1]

        fits.writeto(
            filename=filename,
            data=data,
            header=m.header(),
            overwrite=True,
        )

    def to_hdf(self, filename: str, compress: bool = True):
        compression_kwargs = {"compression": "gzip", "compression_opts": 9} if compress else {}

        with h5py.File(filename, "w") as f:
            f.create_dataset("data", dtype=np.float32, data=self.data, **compression_kwargs)
            if not (self.weight == 1).all().compute():
                f.create_dataset("weight", dtype=np.float32, data=self.weight, **compression_kwargs)
            for dim in ["stokes", "t", "z"]:
                if dim in self.dims:
                    f.create_dataset(dim, data=getattr(self, dim))
            if "nu" in self.dims:
                f.create_dataset("nu", data=self.nu.Hz)
            f.create_dataset("center", dtype=float, data=(self.center[0].deg, self.center[1].deg))
            f.create_dataset("x_res", dtype=float, data=self.x_res.deg)
            f.create_dataset("y_res", dtype=float, data=self.y_res.deg)
            f.create_dataset("units", data=self.units)
            f.create_dataset("frame", data=self.frame.name)
            f.create_dataset("beam", data=self.beam.deg)



from __future__ import annotations

import numpy as np
from maria import Simulation


def test_white_noise_levels():
    sim = Simulation(
        instrument="MUSTANG-2",
        plans="ten_second_zenith_stare",
        site="green_bank",
        noise=True,
    )
    tod = sim.run()[0].to("pW")

    target_error = sim.instrument.bands[0].NEP.to("pW√s") / np.sqrt(sim.plans.duration.s)

    # this is should be distributed as a zero-mean unit-variance Gaussian
    scaled_residuals = tod.noise.compute().mean(axis=1) / target_error

    min_rel_noise = 0.8
    max_rel_noise = 1.2
    if scaled_residuals.std() < min_rel_noise:
        raise RuntimeError(
            f"Noise residuals are too low ({scaled_residuals.std():.03f} < {min_rel_noise})",
        )

    if scaled_residuals.std() > max_rel_noise:
        raise RuntimeError(
            f"Noise residuals are too high ({scaled_residuals.std():.03f} > {max_rel_noise})",
        )



import logging

import numpy as np
import pandas as pd

from maria.utils import compute_diameter, get_rotation_matrix_2d

logger = logging.getLogger("maria")

SHAPES = ["triangle", "square", "hexagon", "octagon", "circle", "rhombus"]
PACKINGS = ["triangular", "square", "sunflower"]


def generate_sunflower_packing(n: int):
    i = np.arange(n)
    golden_angle = np.pi * (3.0 - np.sqrt(5.0))
    x = 0.5966 * np.sqrt(i) * np.cos(golden_angle * i)
    y = 0.5966 * np.sqrt(i) * np.sin(golden_angle * i)

    return pd.DataFrame({"x": x, "y": y})


def generate_square_packing(n_row: int, n_col: int):
    x_side = np.arange(n_col, dtype=float)
    y_side = np.arange(n_row, dtype=float)
    col, row = np.meshgrid(x_side, y_side)

    x = col - n_col // 2 + (n_col + 1) % 2
    y = row - n_row // 2 + (n_row + 1) % 2

    df = pd.DataFrame({"x": x.ravel(), "y": y.ravel(), "row": row.ravel(), "col": col.ravel()})  # noqa
    df = df.sort_values(["row", "col"], ascending=[False, True])  # noqa
    df.index = np.arange(len(df))

    return df


def generate_triangular_packing(n_col: int, n_row: int):
    x_side = np.arange(n_col, dtype=float)
    y_side = np.arange(n_row, dtype=float)
    col, row = np.meshgrid(x_side, y_side)

    x = col - n_col // 2 + (n_col + 1) % 2
    y = row - n_row // 2 + (n_row + 1) % 2 - 0.5 * x
    x *= np.sqrt(3) / 2

    df = pd.DataFrame({"x": x.ravel(), "y": y.ravel(), "row": row.ravel(), "col": col.ravel()})  # noqa
    df = df.sort_values(["row", "col"], ascending=[False, True])  # noqa
    df.index = np.arange(len(df))

    return df


def scaled_distance(x: float, y: float, shape: str, height_scale: float = 1.0):
    NGONS = {"triangle": 3, "square": 4, "hexagon": 6, "octagon": 8, "circle": 1024}

    r = np.sqrt(x**2 + (y / height_scale) ** 2)
    p = np.arctan2(y / height_scale, x)

    if shape in NGONS:
        n_sides = NGONS[shape]
        d = r * np.cos(np.arcsin(np.sin(n_sides / 2 * p)) * 2 / n_sides)

    elif shape == "rhombus":
        d = r * (np.abs(np.cos(p)) / np.sqrt(3) + np.abs(np.sin(p)))

    else:
        raise ValueError()

    return d + 1e-3 * r.max() * p


def generate_2d_pattern(
    n: int = None,
    n_col: int = None,
    n_row: int = None,
    max_diameter: float = None,
    spacing: float = None,
    shape: str = "hexagon",
    rotation: float = 0,
    packing: str = "triangular",
    height_scale: float = 1.0,
    max_iterations: int = 16,
    tol: float = 1e-2,
):
    """
    Generates an [x,y] array of 2D points to serve as an array.
    By convention, the generated points have a nearest-neighbor distance of 1.
    """

    if packing not in PACKINGS:
        raise ValueError(f"Supported array packings are {PACKINGS}.")

    if shape not in SHAPES:
        raise ValueError(f"Supported array shapes are {SHAPES}.")

    # if no detector is supplied, we'll use an iterative method to find the right number.
    # a good first guess might be n = (max_diameter / spacing) ** 2
    n_explicit = (n is not None) or ((n_col is not None) and (n_row is not None))
    n_args = sum([n_explicit, spacing is not None, max_diameter is not None])

    if n_args < 2:
        raise ValueError()

    if not n_explicit:
        iteration = 0
        current_max_diameter = max_diameter / 1e6
        current_n = 3

        # if only the field of view is supplied
        while (iteration < max_iterations) and (np.abs(np.log(current_max_diameter / max_diameter)) > tol):
            offsets = generate_2d_pattern(
                n=current_n,
                spacing=spacing,
                shape=shape,
                rotation=rotation,
                packing=packing,
            )

            current_max_diameter = compute_diameter(offsets)

            logger.debug(
                f"Array generation iteration {iteration + 1} / {max_iterations} (n={current_n}, "
                f"current_diameter={current_max_diameter:.03e}), target_diameter={max_diameter:.03e})."
            )

            n_adjust_factor = max(1e-2, min((max_diameter / current_max_diameter) ** 2, 1e2))
            current_n = int(np.maximum(3, current_n * n_adjust_factor))

            if current_n > 1e6:
                raise RuntimeError()

            iteration += 1

            # print(iteration, current_n, current_max_diameter)

        return offsets

    else:
        if (n is not None) ^ ((n_col is not None) and (n_row is not None)):
            if n is not None:
                n_col = int(2 * np.sqrt(n))
                n_row = int(2 * np.sqrt(n))
        else:
            raise ValueError()

        if packing == "square":
            df = generate_square_packing(n_col=n_col, n_row=n_row)

        elif packing == "triangular":
            df = generate_triangular_packing(n_col=n_col, n_row=n_row)

        elif packing == "sunflower":
            df = generate_sunflower_packing(n=max(n_col, n_row) ** 2)

        if n is None:
            x = 2 * np.abs(df.x) - 0.25
            y = 2 * np.abs(df.y) - 0.25
            subset_index = np.where((x <= n_col) & (y < n_row))[0]

        if n is not None:
            loss = scaled_distance(x=df.x.values, y=df.y.values, shape=shape, height_scale=height_scale)
            subset_index = sorted(np.argsort(loss)[:n])
            df = df.iloc[subset_index]

        X = (get_rotation_matrix_2d(rotation) @ np.stack([df.x.values, df.y.values])).T

        if max_diameter:
            return max_diameter * X / compute_diameter(X)
        return spacing * X



from __future__ import annotations

from . import signal, tod  # noqa
from .processing import process_tod  # noqa
from .tod import TOD, TOD_QUANTITIES  # noqa

TOD.process = process_tod  # to avoid circular imports



from __future__ import annotations

import numpy as np

from ..units import Quantity
from .coordinates import Coordinates, frames  # noqa
from .frame import FRAMES, Frame  # noqa
from .transforms import (  # noqa
    get_center_phi_theta,
    offsets_to_phi_theta,
    phi_theta_to_offsets,
    phi_theta_to_xyz,
    unjitted_offsets_to_phi_theta,
    xyz_to_phi_theta,
)


def infer_center_width_height(coords_list: list, frame: str = "ra_dec", center: tuple[Quantity] = None, square: bool = True):
    frame = Frame(frame)

    if center is None:
        center_phis, center_thetas = [], []
        for coords in coords_list:
            coords_center = coords.center(frame=frame)
            center_phis.append(coords_center[0].rad)
            center_thetas.append(coords_center[1].rad)

        center = get_center_phi_theta(center_phis, center_thetas)

    hull_points_list = []
    for coords in coords_list:
        hull_points_list.append(coords.hull(center=center, frame=frame))

    hull_points = np.concat(hull_points_list, axis=0)

    xmin, ymin = hull_points.min(axis=0)
    xmax, ymax = hull_points.max(axis=0)

    # corners = np.array([[xmin, ymin], [xmin, ymax], [xmax, ymax], [xmax, ymin]])
    # pt = unjitted_offsets_to_phi_theta(corners, *approx_center)
    # total_center = get_center_phi_theta(pt[..., 0], pt[..., 1])

    width, height = 2 * max(-xmin, xmax), 2 * max(-ymin, ymax)

    if square:
        width = height = max(width, height)

    return center, width, height



from __future__ import annotations

import maria
import matplotlib.pyplot as plt
import numpy as np
import pytest
from maria.instrument import Band
from maria.io import fetch
from maria.map import all_maps
from maria.mappers import BinMapper

plt.close("all")


@pytest.mark.parametrize(
    "map_path",
    all_maps,
)  # noqa
def test_maps(map_path):
    m = maria.map.load(fetch(map_path))
    if "nu" not in m.dims:
        m = m.unsqueeze("nu", 150e9)
    assert np.allclose(m.to("K_RJ").to("Jy/pixel").to(m.units).data, m.data).compute()

    m.to("cK_RJ").to_hdf("/tmp/test_write_map.h5")
    new_m_hdf = maria.map.load("/tmp/test_write_map.h5").to(m.units)  # noqa

    assert np.allclose(new_m_hdf.data, m.data)
    assert np.allclose(new_m_hdf.resolution.arcsec, m.resolution.arcsec)

    if "fits" in map_path:
        m.to("cK_RJ").to_fits("/tmp/test_write_map.fits")
        new_m_fits = maria.map.load("/tmp/test_write_map.fits").to(m.units)  # noqa

        assert np.allclose(new_m_fits.data, m.data)
        assert np.allclose(new_m_fits.resolution.arcsec, m.resolution.arcsec)

    m.plot()

    plt.close("all")


def test_map_operations():
    map_filename = fetch("maps/cluster1.fits")

    m1 = maria.map.load(filename=map_filename, nu=90e9)
    m2 = maria.map.load(filename=map_filename, nu=150e9)
    m3 = maria.map.load(filename=map_filename, nu=220e9)

    m4 = m1.extend([m2, m3], dim="nu").unsqueeze("stokes")
    m5, m6 = m4.copy(), m4.copy()
    m5.stokes = "Q"
    m6.stokes = "U"

    m4.extend([m5, m6], dim="stokes")


def test_trivial_recover_original_map():
    f090 = Band(center=90e9, width=30e9)
    f150 = Band(center=150e9, width=40e9)
    f220 = Band(center=220e9, width=50e9)

    map_filename = fetch("maps/cluster1.fits")
    input_map = maria.map.load(filename=map_filename, nu=150e9).to("K_RJ")[..., ::8, ::8]

    input_map.data -= input_map.data.mean().compute()
    map_width = input_map.width.degrees

    array = {
        "field_of_view": map_width / 2,
        "primary_size": 1000,
        "n": 300,
        "bands": [f090, f150, f220],
    }

    instrument = maria.get_instrument(array=array)

    planner = maria.Planner(target=input_map, site="cerro_toco", constraints={"el": (40, 90)})

    plans = planner.generate_plans(
        total_duration=60, sample_rate=50, scan_pattern="daisy", scan_options={"radius": input_map.width.deg / 3}
    )

    sim = maria.Simulation(
        instrument,
        plans=plans,
        map=input_map,
        site="cerro_toco",
        noise=False,
    )

    tods = sim.run()

    mapper = BinMapper(
        center=input_map.center,
        frame=input_map.frame,
        width=input_map.width,
        resolution=input_map.x_res,
        degrees=False,
        tods=tods,
    )

    mapper.add_tods(tods)
    output_map = mapper.run()

    m0 = input_map.data[0, 0, 0].compute()
    m1 = output_map.data[0, :].compute()
    w = output_map.weight[-1, 0].compute()

    relsqres = np.sqrt(np.nansum(w * (m1 - m0) ** 2, axis=(-1, -2)) / np.nansum(w))

    assert all(relsqres < 1e-3)


def test_time_ordered_map_sim():
    time_evolving_sun_path = fetch("maps/sun.h5")
    input_map = maria.map.load(filename=time_evolving_sun_path, nu=100e9, t=1.7e9 + np.linspace(0, 180, 16))
    plan = maria.Plan.generate(
        start_time=1.7e9,
        duration=180,
        scan_center=(input_map.center),
        scan_options={"radius": 0.25},
    )
    sim = maria.Simulation(instrument="test/1deg", site="cerro_toco", plans=plan, map=input_map)
    tods = sim.run()
    tods[0].to("K_RJ").plot()

    plt.close("all")



from __future__ import annotations

import dask.array as da
import numpy as np
import scipy as sp

from .. import utils
from ..beam import construct_beam_filter, separably_filter_2d
from ..coords import Coordinates, get_center_phi_theta
from ..functions import approximate_normalized_matern
from ..instrument import Instrument
from ..weather import Weather

MIN_SAMPLES_PER_RIBBON = 2
RIBBON_SAMPLE_DECAY = 2
JITTER_LEVEL = 1e-4

matern_callback = approximate_normalized_matern


class TurbulentLayer:
    """
    The simplest possible atmospheric model.
    This model is only appropriate for single instruments, i.e. when the baseline is zero.
    """

    def __init__(
        self,
        instrument: Instrument,
        boresight: Coordinates,
        weather: Weather,
        depth: float,
        res: float,
        turbulent_outer_scale: float = 500,
        verbose: bool = False,
        timestep: float = 0.05,
        **kwargs,
    ):
        self.instrument = instrument
        self.boresight = boresight.downsample(timestep=timestep)
        self.weather = weather
        self.depth = depth
        self.res = res
        self.timestep = timestep

        self.sim_time = self.boresight.time
        self.sim_az = self.boresight.az
        self.sim_el = self.boresight.el

        # this is approximately correct
        # self.depth = self.depth / np.sin(np.mean(self.plan.el))

        # this might change
        self.angular_outer_scale = turbulent_outer_scale / self.depth

        # returns the beam fwhm for each detector at the layer distance
        # self.physical_beam_fwhm = self.instrument.physical_fwhm(self.depth)
        # self.angular_beam_fwhm = self.physical_beam_fwhm / self.depth

        self.angular_resolution = self.res / self.depth

        if verbose:
            print(f"{self.angular_resolution=}")

        self.layer_altitude = self.weather.altitude + self.depth / np.sin(self.sim_el)

        layer_wind_north = sp.interpolate.interp1d(
            self.weather.altitude,
            self.weather.wind_north,
            axis=0,
        )(self.layer_altitude)
        layer_wind_east = sp.interpolate.interp1d(
            self.weather.altitude,
            self.weather.wind_east,
            axis=0,
        )(self.layer_altitude)

        angular_velocity_x = (+layer_wind_east * np.cos(self.sim_az) - layer_wind_north * np.sin(self.sim_az)) / self.depth

        angular_velocity_y = (-layer_wind_east * np.sin(self.sim_az) + layer_wind_north * np.cos(self.sim_az)) / self.depth

        if verbose:
            print(f"{(layer_wind_east, layer_wind_north)=}")

        # compute the offset with respect to the center of the scan
        center_az, center_el = get_center_phi_theta(self.sim_az, self.sim_el)
        offsets = self.boresight.offsets(frame="az/el", center=(center_az, center_el))

        # the angular position of the boresight over time WRT the atmosphere
        # this has dimensions (2, time index)
        self.boresight_angular_position = np.c_[
            offsets[..., 0] + np.cumsum(angular_velocity_x * np.gradient(self.sim_time)),
            offsets[..., 1] + np.cumsum(angular_velocity_y * np.gradient(self.sim_time)),
        ]

        # find the detector offsets which form a convex hull
        self.detector_offsets = np.radians(
            np.c_[self.instrument.sky_x, self.instrument.sky_y],
        )

        # add a small circle of offsets to account for the beams
        unit_circle_complex = np.exp(1j * np.linspace(0, 2 * np.pi, 64 + 1)[:-1])
        unit_circle_offsets = np.c_[
            np.real(unit_circle_complex),
            np.imag(unit_circle_complex),
        ]

        # this is a convex hull for the instrument if it's staring
        stare_convex_hull = sp.spatial.ConvexHull(
            (
                self.detector_offsets[None, :, None]
                + self.instrument.angular_fwhm(depth)[:, None, None] * unit_circle_offsets[None]
            ).reshape(-1, 2),
        )
        stare_convex_hull_points = stare_convex_hull.points[stare_convex_hull.vertices].reshape(-1, 2)

        # convex hull downsample index, to get to 1 second
        chds_index = [
            *np.arange(
                0,
                len(self.sim_time),
                int(np.maximum(np.median(np.diff(self.sim_time)), 1)),
            ),
            -1,
        ]

        # this is a convex hull for the atmosphere
        atmosphere_hull = sp.spatial.ConvexHull(
            (self.boresight_angular_position[chds_index, None] + stare_convex_hull_points[None]).reshape(-1, 2),
        )
        self.atmosphere_hull_points = atmosphere_hull.points.reshape(-1, 2)[atmosphere_hull.vertices]

        # R takes us from the real (dx, dy) to a more compact (cross_section, extrusion) frame
        self.optres = utils.linalg.optimize_area_minimizing_rotation_matrix(
            self.atmosphere_hull_points,
        )

        assert self.optres.success
        self.R = utils.linalg.get_rotation_matrix_2d(self.optres.x[0])

        #
        #          ^      xxxxxxxxxxxx
        #          |      xxxxxxxxxxxx
        #   cross-section xxxxxxxxxxxx
        #          |      xxxxxxxxxxxx
        #          @      xxxxxxxxxxxx
        #

        trans_points = self.atmosphere_hull_points @ self.R.T
        cross_section_min = trans_points[:, 0].min()
        cross_section_max = trans_points[:, 0].max()
        extrusion_min = trans_points[:, 1].min()
        extrusion_max = trans_points[:, 1].max()

        self.cross_section_side = np.arange(
            cross_section_min,
            cross_section_max + self.angular_resolution,
            self.angular_resolution,
        )
        self.extrusion_side = np.arange(
            extrusion_min,
            extrusion_max,
            self.angular_resolution,
        )

        self.n_cross_section = len(self.cross_section_side)
        self.n_extrusion = len(self.extrusion_side)

        CROSS_SECTION, EXTRUSION = np.meshgrid(
            self.cross_section_side,
            self.extrusion_side,
        )

        self.TRANS_POINTS = np.concatenate(
            [CROSS_SECTION[..., None], EXTRUSION[..., None]],
            axis=-1,
        )

        extrusion_indices = [
            0,
            *(2 ** np.arange(0, np.log(self.n_extrusion) / np.log(2))).astype(int),
            self.n_extrusion - 1,
        ]

        extrusion_sample_index = []
        cross_section_sample_index = []
        for i, extrusion_index in enumerate(extrusion_indices):
            n_ribbon_samples = np.minimum(
                np.maximum(
                    int(self.n_cross_section * RIBBON_SAMPLE_DECAY**-i),
                    MIN_SAMPLES_PER_RIBBON,
                ),
                self.n_cross_section,
            )
            cross_section_indices = np.unique(
                np.linspace(0, self.n_cross_section - 1, n_ribbon_samples).astype(int),
            )
            cross_section_sample_index.extend(cross_section_indices)
            extrusion_sample_index.extend(
                np.repeat(extrusion_index, len(cross_section_indices)),
            )

        self.extrusion_sample_index = np.array(extrusion_sample_index)
        self.cross_section_sample_index = np.array(cross_section_sample_index)

        live_edge_positions = np.c_[
            self.cross_section_side,
            np.repeat(extrusion_min - self.angular_resolution, self.n_cross_section),
        ]

        sample_positions = np.c_[
            CROSS_SECTION[self.extrusion_sample_index, self.cross_section_sample_index],
            EXTRUSION[self.extrusion_sample_index, self.cross_section_sample_index],
        ]

        # the sampling index will look something like:
        #
        #          x111010000000001 ...
        #          x100000000000000 ...
        #          x110000000000000 ...
        #          x100000000000000 ...
        #  leading x111000000000001 ...
        #  edge    x100000000000000 ...
        #          x110000000000000 ...
        #          x100000000000000 ...
        #          x111000000000001 ...
        #

        self.n_live_edge = len(live_edge_positions)
        self.n_sample = len(sample_positions)

        if verbose:
            print(f"{self.n_extrusion=}")
            print(f"{self.n_live_edge=}")
            print(f"{self.n_sample=}")

        # sample upper {i,j}
        i, j = np.triu_indices(self.n_sample, k=1)
        COV_S_S = np.eye(self.n_sample) + JITTER_LEVEL
        COV_S_S[i, j] = matern_callback(
            np.sqrt(np.square(sample_positions[j] - sample_positions[i]).sum(axis=1)) / self.angular_outer_scale,
            5 / 6,
            n_test_points=256,
        )
        COV_S_S[j, i] = COV_S_S[i, j]

        # this one is explicit
        COV_LE_S = matern_callback(
            np.sqrt(
                np.square(sample_positions[None] - live_edge_positions[:, None]).sum(
                    axis=2,
                ),
            )
            / self.angular_outer_scale,
            5 / 6,
            n_test_points=256,
        )

        # live edge upper {i,j}
        i, j = np.triu_indices(self.n_live_edge, k=1)
        COV_LE_LE = np.eye(self.n_live_edge) + JITTER_LEVEL
        COV_LE_LE[i, j] = matern_callback(
            np.sqrt(
                np.square(live_edge_positions[j] - live_edge_positions[i]).sum(axis=1),
            )
            / self.angular_outer_scale,
            5 / 6,
            n_test_points=256,
        )
        COV_LE_LE[j, i] = COV_LE_LE[i, j]

        # this is typically the bottleneck
        inv_COV_S_S = utils.linalg.fast_psd_inverse(COV_S_S)

        # compute the weights
        self.A = COV_LE_S @ inv_COV_S_S
        self.B = np.linalg.cholesky(COV_LE_LE - self.A @ COV_LE_S.T)
        self.shaped_values = np.zeros(
            (self.n_extrusion, self.n_cross_section),
            dtype=np.float32,
        )

        self.shaped_values = da.asarray(self.shaped_values)

        self.atmosphere_detector_points = (self.detector_offsets[:, None] + self.boresight_angular_position[None]) @ self.R.T

    def generate(self):
        n_steps = self.n_extrusion

        extruded_values = utils.linalg.extrude(
            values=self.shaped_values.ravel(),
            A=self.A,
            B=self.B,
            n_steps=n_steps,
            n_i=self.n_extrusion,
            n_j=self.n_cross_section,
            i_sample_index=self.extrusion_sample_index,
            j_sample_index=self.cross_section_sample_index,
        )

        self.shaped_values = extruded_values.reshape(
            self.n_extrusion,
            self.n_cross_section,
        )

    def sample(self):
        detector_values = np.zeros(self.atmosphere_detector_points.shape[:-1])
        for band in self.instrument.bands:
            # we assume the atmosphere looks the same for every nu in the band

            band_index = self.instrument.dets.subset(band_name=band.name).index
            band_angular_fwhm = self.instrument.angular_fwhm(z=self.depth)[band_index].mean()

            F = construct_beam_filter(
                fwhm=band_angular_fwhm,
                res=self.angular_resolution,
                beam_profile=self.instrument.beam_profile,
            )
            FILTERED_VALUES = separably_filter_2d(self.shaped_values, F)
            FILTERED_VALUES = self.shaped_values

            detector_values[band_index] = sp.interpolate.RegularGridInterpolator(
                (self.cross_section_side, self.extrusion_side),
                FILTERED_VALUES.T,
            )(self.atmosphere_detector_points[band_index])

        return detector_values

    # def simulate_integrated_water_vapor(self):
    #     detector_values = self.simulate_normalized_effective_water_vapor()

    #     # this is "zenith-scaled"
    #     self.line_of_sight_pwv = (
    #         self.weather.pwv
    #         * (1.0 + self.pwv_rms_frac * detector_values)
    #         / np.sin(self.EL)
    #     )



import numpy as np


def dms_to_deg(d: float = 0, m: float = 0, s: float = 0):
    return np.radians(d + m / 60 + s / 3600)


def hms_to_deg(h: float = 0, m: float = 0, s: float = 0):
    return np.radians(h * 15 + m * 15 / 60 + s * 15 / 3600)


def deg_to_signed_dms(x: float, precision: int = 6):
    x = round(x, precision)
    sign = -1 if x < 0 else 1
    mnt, sec = divmod(abs(x) * 3600, 60)
    deg, mnt = divmod(mnt, 60)
    return int(sign), int(deg), int(mnt), sec


def deg_to_signed_hms(x: float, precision: int = 6):
    x = round(x, precision)
    sign = -1 if x < 0 else 1
    mnt, sec = divmod(abs(x) * 3600 / 15, 60)
    deg, mnt = divmod(mnt, 60)
    return int(sign), int(deg), int(mnt), sec



class NoSuitablePlansError(Exception):
    def __init__(self, constraints, max_chunk_duration, tmin, tmax):
        msg = f"No chunks of duration {max_chunk_duration} satisfy the constraints {constraints} between {tmin} and {tmax}"
        super().__init__(msg)
